/*
                 ,----------------, 
           ,-----------------------,
         ,"                      ,"|
        +-----------------------+  |
        |  .-----------------.  |  |
        |  |                 |  |  | 
        |  | Hack the Gibson |  |  |
        |- |      TERM       |  |  |
        |- |                 |  |  | 
        |- |                 |  |  | 
        |- `-----------------'  |,"
        +-----------------------+
           /_)______________(_/
      _______________________________ 
     /  oooooooooooooooo  .o.  oooo /|
    / ==ooooooooooooooo==.o.  ooo= // 
   /_==__==========__==_ooo__ooo=_/'
   `-----------------------------'

  Hack the Gibson TERM is GNU software
   please read ../COPYING for details

hgterm by Thea DeSilva
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <GL/gl.h>
#include <SDL.h>
#include <SDL_image.h>

#include "include.h"

extern float logo_rotate;

b_back_t back1, back2, back3, back4;
int cur_mesh;
float norm_neb[12];
 float bell_alpha, bell_alpha2;
 int bell_dir, bell_dir2;

float bell_x, bell_y;
float find_place;

char *slide_show[] = {
 "/var/q.jpeg",
 "/var/q2.jpeg",
 "/var/q3.jpeg",
 "/var/q4.jpeg"
};

struct {
 int current_slide1;
 int current_slide2;
 int slide_count;
 int cached_slides[2];
} slides;

/* 
 * copy a box region into another buffer
 */
unsigned char *_b_region2rgb(unsigned char *src, int src_width, int x, int y, int w, int h) {
 int i,j;
 unsigned char *dat;
 dat = (unsigned char *)malloc( w * h * 3 );
 for(i=0;i<h;i++)
  for(j=0;j<w;j++) {
   dat[ T2L(j,i,w)*3   ]= src[  T2L(j+x,i+y,src_width)*3   ];
   dat[(T2L(j,i,w)*3)+1]= src[ (T2L(j+x,i+y,src_width)*3)+1]; 
   dat[(T2L(j,i,w)*3)+2]= src[ (T2L(j+x,i+y,src_width)*3)+2];
  }
 return dat;
}

/* Load the texture(s) for a background */

int _b_freetex(b_back_t *back) {
 int i;
 i = (back->x_num+(back->x_pad !=0 ? 1 : 0)) *  
     (back->y_num+(back->y_pad !=0 ? 1 : 0));
 glDeleteTextures( i, back->i_tid);
}

int _b_loadtex(char *filename, b_back_t *back) {
  SDL_Surface *image;
  unsigned char *dat, *dat2;
  int c, x, y, i;
  image = file2surface(filename);
  if(!image) {
   e_push("file2surface"); 
   return 0;
  }
 dat = surface2rgb(image, &c);
 if(!dat) { e_push("surface2rgb"); return 0; }

 back->x_num = (int)(image->w / 256);
 back->y_num = (int)(image->h / 256);
 back->x_pad = (int)(image->w % 256);
 back->y_pad = (int)(image->h % 256);

 back->i_tid = (int *)malloc( (back->x_num+(back->x_pad !=0 ? 1 : 0))* 
		              (back->y_num+(back->y_pad !=0 ? 1 : 0))*
	                       sizeof(int));
 i = 0;
 
 /* Upload the textures to opengl */

 for(y=0; y< back->y_num + ( back->y_pad != 0 ? 1 : 0); y++) 
  for(x=0;x< back->x_num + ( back->x_pad != 0 ? 1 : 0); x++) {
   if(x>back->x_num-1) 
    if(y>back->y_num-1) {
     dat2 = _b_region2rgb(dat, image->w, x*256,y*256,back->x_pad,
		                                     back->y_pad);
     t_mem2tex(dat2,back->x_pad, back->y_pad, &back->i_tid[i]);
     i++;
     free(dat2);
    } else {
     dat2 = _b_region2rgb(dat, image->w, x*256,y*256,back->x_pad,256);
     t_mem2tex(dat2,back->x_pad,256,&back->i_tid[i]);
     i++;
     free(dat2);
    }
   else 
    if(y>back->y_num-1) {
     dat2 = _b_region2rgb(dat, image->w, x*256,y*256,256,back->y_pad);
     t_mem2tex(dat2,256, back->y_pad, &back->i_tid[i]);
     i++;
     free(dat2);
    } else {
     dat2 = _b_region2rgb(dat, image->w, x*256, y*256, 256, 256);
     t_mem2tex(dat2,256,256, &back->i_tid[i]);
     i++;
     free(dat2);
    }
  }
 if(c==1) free(dat);
 SDL_FreeSurface(image);
 return 1;
}

int b_init(void) {
 bell_x = 0;
 bell_y = 0;
 find_place = -233004;
 bell_alpha = 0.0f;
 bell_alpha2 = 0.0f;
 bell_dir = 0;
 bell_dir2 = 0;
 slides.current_slide1 = 0;
 slides.current_slide2 = 1;
 slides.cached_slides[0] = 0;
 slides.cached_slides[1] = 1;
 _b_loadtex(slide_show[slides.current_slide1],  &back1);
 _b_loadtex(slide_show[slides.current_slide2],  &back2);
 
 _b_loadtex(slide_show[2],  &back3);
 _b_loadtex(slide_show[3],  &back4);

}

void _b_draw_pix(b_back_t *back, float alpha, int x1, int y1, int x2, int y2) {
 int x, y, i;
 float xs, ys;
 i = 0;	

 glEnable(GL_BLEND);
 glDisable(GL_DEPTH_TEST);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE);
 glEnable(GL_TEXTURE_2D);
 glShadeModel(GL_FLAT);

 /*
 xs = (((float)(x2 - x1) / (255.0 * (float)((back->x_num*MAX_TEXX) + back->x_pad))) * (float)MAX_TEXX) * (float)MAX_TEXX;
 ys = (((float)(y2 - y1) / (255.0 * (float)((back->y_num*MAX_TEXY) + back->y_pad))) * (float)MAX_TEXY) * (float)MAX_TEXY;
*/
 xs = 256;
 ys = 256;
 
 
 for(y=0; y< back->y_num+(back->y_pad != 0 ? 1 : 0); y++)
  for(x=0;x< back->x_num+(back->x_pad != 0 ? 1 : 0); x++) {
   glBindTexture(GL_TEXTURE_2D, back->i_tid[i]);
   i++;
   glBegin(GL_QUADS);
   glColor4f(1.0f, 1.0f, 1.0f, alpha);
    glTexCoord2f( 0.0f, 0.0f); glVertex3f( 0.0f + ( x * 256),   0.0f + ( y * 256), 0.0f);
    glTexCoord2f( 0.9961f, 0.0f); glVertex3f( xs   + ( x * 256),   0.0f + ( y * 256), 0.0f);
    glTexCoord2f( 0.9961f, 0.9961f); glVertex3f( xs   +( x * 256),    ys   + ( y * 256), 0.0f);
    glTexCoord2f( 0.0f, 0.9961f); glVertex3f( 0.0f + ( x * 256),   ys + ( y * 256), 0.0f);
   glEnd();
  }
 glShadeModel(GL_SMOOTH);
}

int b_disp(void) {
  int x, y;
 unsigned char dat; 
// _b_draw_pix( &back2, bell_alpha, 0 - (bell_alpha2 * 100), 0 - (bell_alpha2 * 100), d_width+ (bell_alpha2 * 100), d_height+ (bell_alpha2 * 100));
 // 1.0f - bell_alpha
 _b_draw_pix( &back1, 1.0f, 0, 0, d_width, d_height);
 
//  _b_draw_pix( &back1, 1.0f, 0,0, d_width, d_height);
 
}

void b_tick(void) {
 logo_rotate+=0.016f;
}
