/*
                 ,----------------, 
           ,-----------------------,
         ,"                      ,"|
        +-----------------------+  |
        |  .-----------------.  |  |
        |  |                 |  |  | 
        |  | Hack the Gibson |  |  |
        |- |      TERM       |  |  |
        |- |                 |  |  | 
        |- |                 |  |  | 
        |- `-----------------'  |,"
        +-----------------------+
           /_)______________(_/
      _______________________________ 
     /  oooooooooooooooo  .o.  oooo /|
    / ==ooooooooooooooo==.o.  ooo= // 
   /_==__==========__==_ooo__ooo=_/'
   `-----------------------------'

  Hack the Gibson TERM is GNU software
   please read ../COPYING for details

hgterm by Thea DeSilva
*/


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <GL/gl.h>
#include <SDL.h>
#include <SDL_image.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>

#include "include.h"


// 116 x 48

float logo_rotate;

int cur_frame;

unsigned char *key_buf;
int key_rep, key_del;
int key_pos;
int key_state;
int key_down, key_down_state;

int cursor_blink, cblink;

int shift_count, norm_count;
key_ent *shift_keys;
key_ent *norm_keys;

int d_width, d_height, d_depth;
SDL_Surface *d_screen;



void v_termout( char *s, int len, f_plane_t *plane);

char keymap[128][3] = { /* ascii for syms under [norm|shift|ctrl] */
#include "keymap.h"
};

void d_screen_shot(char *filename) {
 system("xwd -root -out /tmp/q.xwd");
}

/* Set window size */

void d_resize(int x, int y) {
 d_screen = SDL_SetVideoMode(x,y,d_depth, SDL_OPENGL|SDL_RESIZABLE);
 glViewport(0,0, (GLint)x, (GLint)y);
 d_width = x;
 d_height = y; 
}

/* Set up gl matrix for screen space 
 * e.g. Upper left hand corner is 0,0
 *      lower right is d_width, d_height
 */

void d_ss(void) {
 glLoadIdentity();
 glScalef( 1.0f / (d_width/2), 
	   -1.0f / (d_height/2), 1.0f); 
// glRotatef(-90.0f, 0.0f, 0.0f, 1.0f);
 glTranslatef( (float)(-1* (d_width/2)), 
	       (float)(-1* (d_height/2)), 0.0f);
}

/* Set up gl matrix for world space
 * e.g. Upper left hand corner is -512, -512
 *      lower right is 512, 512
 */

void d_ws(void) {
 glLoadIdentity();
 glScalef( 1.0f/1024.0f, -1.0f/1024.0f, 1.0f);
}

/* Initialize everything about the display */

int d_init(int x, int y, int d) {
 SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_NOPARACHUTE);
 d_screen = SDL_SetVideoMode(x,y,d,SDL_OPENGL|SDL_RESIZABLE);
 if(!d_screen) { 
  e_push("SDL_SetVideoMode");
  return 0;
 }
 glShadeModel(GL_SMOOTH);
 glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
 glClearDepth(0.0f);
 glDepthFunc(GL_LEQUAL);
 glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//  glDisable(GL_DITHER);
 d_depth = d;
 d_resize(x,y);
 d_ss();
 key_buf = (unsigned char *)malloc(512);
 key_pos = -1;
 key_state = 0;
 key_rep = WAIT;
 key_del = WAIT;
 key_rep = 0;
 key_del = 0;

 SDL_EnableKeyRepeat( 1000, 25);

 logo_rotate = 0.0f;
 
 return 1;
}

void d_draw_texture_on_angle( float ang, float *dx1, float *dx2, float *dy1, float *dy2) {
 float cos_, sin_;
    cos_ = cos(ang);
    sin_ = sin(ang); 

    *dx1 = ((d_width/4) *3) + (( -150.0f * cos_) - (-150.0f * sin_));
    *dy1 = ((d_height/4)*3) + (( -150.0f * cos_) + (-150.0f * sin_));

    *dx2 = ((d_width/4) *3) + ((  150.0f * cos_) - ( 150.0f * sin_));
    *dy2 = ((d_height/4)*3) + ((  150.0f * cos_) + ( 150.0f * sin_));
 

    glBegin(GL_QUADS);
     glTexCoord2f(0.0f, 0.0f); glVertex3f(*dx1, *dy1, 0.0f);
     glTexCoord2f(1.0f, 0.0f); glVertex3f(*dx2, *dy1, 0.0f);
     glTexCoord2f(1.0f, 1.0f); glVertex3f(*dx2, *dy2, 0.0f);
     glTexCoord2f(0.0f, 1.0f); glVertex3f(*dx1, *dy2, 0.0f);
    glEnd();  
 
}


void d_drawscreen(void) {
   int i; 
   float dx1, dy1, dx2, dy2;
   unsigned char dat;
   if(cursor_blink == GO) { cblink++; cursor_blink = WAIT; }
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 //  b_disp(); 

    glEnable(GL_TEXTURE_2D);
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBindTexture(GL_TEXTURE_2D, t_tid[GRAPHIC]);

    glColor4f(0.0f,1.0f,1.0f,1.0f / 3 );

    d_draw_texture_on_angle( logo_rotate +((M_PI/3)   ), &dx1, &dx2, &dy1, &dy2 );
    d_draw_texture_on_angle( logo_rotate +((M_PI/3) *2), &dx1, &dx2, &dy1, &dy2 );
    d_draw_texture_on_angle( logo_rotate +((M_PI/3) *3), &dx1, &dx2, &dy1, &dy2 );

#ifdef UNDUMP
    // Heyas
#else
   dat = d_pullkey();
   if(dat != KEY_BUF_EMPTY)  {
    if(dat == '\r') dat = '\n';
    write(qfd, &dat, sizeof(unsigned char) );
   }
#endif
   f_draw_planes();
   td_draw_td(tds[0]);
   SDL_GL_SwapBuffers();
   return;
}

void d_bell(void) {
}

unsigned char d_pullkey(void) {
 if( key_pos == -1) return KEY_BUF_EMPTY;
 key_pos--;
 return key_buf[key_pos+1];
}

void d_keypress(int sym, int state) {
 char buf[3];
 if(key_pos==511) { d_bell(); return; }
 switch(sym) {
  case 273:
   vt_send(K_UP, f_screen_plane);
   return;
  case 274:
   vt_send(K_DN, f_screen_plane);
   return;
  case 276:
   vt_send(K_LT, f_screen_plane);
   return;
  case 275:
   vt_send(K_RT, f_screen_plane);
   return;
  default:
   key_buf[key_pos+1] = keymap[sym][state];
   break;
 }
 if(key_buf[key_pos+1] == '\r') make_sound(SD_RET); else make_sound(SD_TICK);
 key_pos++;
}

void d_input(void) {
 SDL_Event event;
 while(SDL_PollEvent(&event) )  {
  switch(event.type) {
   case SDL_VIDEORESIZE:
//    d_resize(event.resize.w, event.resize.h);
/*    d_ss();  */
    break;
   case SDL_QUIT:
    SDL_Quit(); 
    exit(-1);
    break;
   case SDL_VIDEOEXPOSE:
    key_state = 0;
   break;
   case SDL_ACTIVEEVENT:
    if(event.active.state == SDL_APPACTIVE) 
     if(event.active.gain == 0) { 
      s_ssync( d_drawscreen, DISABLE, CALLBACK);
      s_ssync( b_tick,       DISABLE, CALLBACK);
      setpriority( PRIO_PROCESS, 0, 10);
     } else {
      s_ssync( d_drawscreen, ENABLE, CALLBACK);
      s_ssync( b_tick,       ENABLE, CALLBACK);
      setpriority( PRIO_PROCESS, 0, 0);
     }
    break;
   case SDL_MOUSEMOTION:
   //  bell_x = event.motion.x;
   break;
   case SDL_KEYDOWN:
    if(event.key.keysym.scancode == 37) {
     key_state = 2; 
     break;
    }
    if(event.key.keysym.scancode == 50) {
     key_state = 1;
     break;
    }
    
    key_down =       event.key.keysym.sym;
    key_down_state = key_state;
    
    d_keypress( key_down, key_down_state);
   case SDL_KEYUP:
    if(event.key.keysym.scancode == 50 ||
       event.key.keysym.scancode == 37) {
     key_state = 0;
     break;
    }
    key_down = -1;
    break;
    
  }
 }
 /*
 if(key_rep == GO)
  s_ssync( &key_del, ENABLE, SYNC);
 if(key_del == GO) {
  key_del=WAIT;
  if(key_rep == GO)
   d_keypress( key_down, key_down_state);
 }
 */
}
