/*
                 ,----------------, 
           ,-----------------------,
         ,"                      ,"|
        +-----------------------+  |
        |  .-----------------.  |  |
        |  |                 |  |  | 
        |  | Hack the Gibson |  |  |
        |- |      TERM       |  |  |
        |- |                 |  |  | 
        |- |                 |  |  | 
        |- `-----------------'  |,"
        +-----------------------+
           /_)______________(_/
      _______________________________ 
     /  oooooooooooooooo  .o.  oooo /|
    / ==ooooooooooooooo==.o.  ooo= // 
   /_==__==========__==_ooo__ooo=_/'
   `-----------------------------'

  Hack the Gibson TERM is GNU software
   please read ../COPYING for details

hgterm by Thea DeSilva
*/

#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/extensions/xf86vmode.h>
#include <signal.h>
#include <SDL.h>

#include "include.h"

#define IS_SLEEPING		1
#define IS_NOT_SLEEPING		0

s_callback_t s_callbacks[1024]; /* Call-back table */

/* 
 * Syncs are used as queues for processing time sensitive
 * stuff. Kinda like a wait_v_interrupt or something 
 */

s_sync_t     s_syncs[1024];     /* Sync table */
int s_time_base;		/* Global time base */
int _s_pid;
int _s_int2_running;

/* get the refresh rate of the display in hertz*/

int _s_get_refresh(void) {
 Display *dpy;
 int dotclock;
 int vsync;
 XF86VidModeModeLine ml;

 dpy = XOpenDisplay(NULL);
 XF86VidModeGetModeLine(dpy, DefaultScreen(dpy), &dotclock, &ml);
 vsync = ((dotclock * 1000 / ml.htotal) * 1000 / ml.vtotal) / 1000;

 return vsync;
}

/* I'm not sure if this actually works, but the idea is to take 
 * callbacks, which may or may not take a while to do out of the 
 * primary timer interrupt. If SDL doesn't block further interrupts
 * while processing _s_interrupt, we're cool
 */

void _s_interrupt2(int sig_num) {
 int i;
 _s_int2_running = 1;
 
/* XXX Maintain a linked-list of the callbacks and syncs so that we do not need to walk through the entire table */
 
 for(i=0;i<1024;i++)
  if(s_callbacks[i].callback != 0) {
   if(s_callbacks[i].sleep == IS_SLEEPING) {
    s_callbacks[i].tsleep--;
    if(s_callbacks[i].tsleep == 0)
     s_callbacks[i].sleep = IS_NOT_SLEEPING;   
   } else {
    s_callbacks[i].clock++;
    if(s_callbacks[i].clock == s_callbacks[i].reset) {
     s_callbacks[i].clock = 0;
     if(s_callbacks[i].enabled == ENABLE) { 
      s_callbacks[i].enabled = 0;
      s_callbacks[i].callback();
      s_callbacks[i].enabled = 1;
     }
    }
   }
  }

 signal(SIGUSR1, _s_interrupt2);
 _s_int2_running = 0;
}

int _s_interrupt(int qq) {
 int i;
 for(i=0;i<1024;i++) 
  if(s_syncs[i].reg != 0) {
   if(s_syncs[i].enabled == ENABLE) {
    s_syncs[i].clock++;
    if(s_syncs[i].clock == s_syncs[i].reset) {
     s_syncs[i].clock = 0;
     *s_syncs[i].reg = GO;
    }
   }
  }
  
 if(_s_int2_running) return;
 else kill(_s_pid, SIGUSR1);
 return s_time_base;
}

int s_init(void) {
 _s_pid = getpid();
 signal(SIGUSR1, _s_interrupt2);
 _s_int2_running = 0;
 s_time_base = (1000 / _s_get_refresh()) / 2;
 SDL_SetTimer(s_time_base, (SDL_TimerCallback)_s_interrupt); 
 bzero( s_syncs, 1024 * sizeof( s_sync_t ) );
 bzero( s_callbacks, 1024 * sizeof( s_callback_t ) );
 return 1;
}

int s_register(int tf, void *reg, int type) {
 int empty;
 if( type == CALLBACK) {
  for(empty = 0; empty < 1024; empty++) 
   if( (int)s_callbacks[empty].callback == 0) break;
  s_callbacks[empty].reset  = tf;
  s_callbacks[empty].clock = 0;
  s_callbacks[empty].callback = reg;
  s_callbacks[empty].enabled = DISABLE;
  s_callbacks[empty].sleep = IS_NOT_SLEEPING;
  s_callbacks[empty].tsleep = 0;
  return empty;
 } else {
  for(empty = 0; empty < 1024; empty++)
   if( (int)s_syncs[empty].reg == 0) break;
  s_syncs[empty].reset = tf;
  s_syncs[empty].clock = 0;
  s_syncs[empty].reg = reg;
  *(int *)reg = WAIT;
  s_syncs[empty].enabled = DISABLE;
  return empty;
 }
}

int s_deregister(void *reg, int type) {
 int empty;
 if( type == CALLBACK) {
  for(empty = 0; empty< 1024; empty++)
   if((int)s_callbacks[empty].callback = (int)reg) break;
  s_callbacks[empty].callback = 0;
  s_callbacks[empty].enabled = DISABLE;
 } else {
  for(empty = 0; empty< 1024; empty++)
   if((int)s_syncs[empty].reg == (int)reg) break;
  s_syncs[empty].reg = 0;
  s_syncs[empty].enabled = DISABLE; 
 }
}

int s_sleep( void *reg, int len) {
 int slot;
 for(slot = 0; slot < 1024;slot++)
  if( (int)s_callbacks[slot].callback == (int)reg)
   break;
 s_callbacks[slot].tsleep = len;
 s_callbacks[slot].sleep = IS_SLEEPING;
}

int s_enabled( void *reg, int type) {
 int slot;
 if(type == SYNC) {
  for(slot = 0; slot< 1024; slot++)
   if((int)s_syncs[slot].reg == (int)reg) 
    return s_syncs[slot].enabled;
 } else {
  for(slot = 0; slot< 1024;slot++)
   if( (int)s_callbacks[slot].callback == (int)reg)
    return s_callbacks[slot].enabled;
 }
}

void s_ssync( void *reg, int enable, int type) {
 int slot;
 if( type == CALLBACK) {
  for(slot = 0; slot < 1024; slot++) 
   if((int)s_callbacks[slot].callback == (int)reg) break;
  s_callbacks[slot].enabled = enable;

 } else {
  for(slot = 0; slot < 1024; slot++)
   if( (int)s_syncs[slot].reg == (int)reg) break;
  if(enable == ENABLE) 
   s_syncs[slot].enabled = ENABLE;
  else {
   s_syncs[slot].enabled = DISABLE;
   s_syncs[slot].clock = 0;
  }
 }
}

/* XXX these should be moved into the appropriate ?_init()s */

void s_set(int type) {
 int i;
 if( (type & INIT) > 0) { 
  s_register(1, d_drawscreen, CALLBACK);
  s_register(2,b_tick, CALLBACK);
  s_register(2,d_input, CALLBACK);
//  s_register( TPS(1), &key_rep, SYNC);
//  s_register(TPS(30), &key_del, SYNC);
  s_register(40, &cursor_blink, SYNC);
 }
 if( (type & ENABLE_THEM) > 0) {
  s_ssync( d_drawscreen, ENABLE, CALLBACK);
  s_ssync( b_tick, ENABLE, CALLBACK);
  s_ssync( d_input, ENABLE, CALLBACK);
  s_ssync( &cursor_blink, ENABLE, SYNC);
 }
}
