/*
                 ,----------------, 
           ,-----------------------,
         ,"                      ,"|
        +-----------------------+  |
        |  .-----------------.  |  |
        |  |                 |  |  | 
        |  | Hack the Gibson |  |  |
        |- |      TERM       |  |  |
        |- |                 |  |  | 
        |- |                 |  |  | 
        |- `-----------------'  |,"
        +-----------------------+
           /_)______________(_/
      _______________________________ 
     /  oooooooooooooooo  .o.  oooo /|
    / ==ooooooooooooooo==.o.  ooo= // 
   /_==__==========__==_ooo__ooo=_/'
   `-----------------------------'

  Hack the Gibson TERM is GNU software
   please read ../COPYING for details

hgterm by Thea DeSilva
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <GL/gl.h>
#include <SDL.h>
#include <SDL_image.h>

#include "err.h"
#include "util.h"

vect_t accum_v;
float  accum_f;

/*
 * Given a SDL_Surface convert it to RGB data and return a pointer to 
 * that data, if no convertion was make put 1 into c 
 */

unsigned char *surface2rgb( SDL_Surface *image, int *c) {
 unsigned char *pixels;
 unsigned char *dat;
 int x;
 *c = 0;
 if(image->format->BytesPerPixel != 3) { /* Convert from indexed to RGB */
  if(image->format->BytesPerPixel != 1) {
   e_push("unsupported color type");
   return (unsigned char *)0;
  }
  dat=(unsigned char *)malloc( image->w * image->h * 3);
  pixels = image->pixels;
  for(x=0;x<(image->w*image->h);x++) {
   dat[(x*3)  ] = image->format->palette->colors[pixels[x]].r;
   dat[(x*3)+1] = image->format->palette->colors[pixels[x]].g;
   dat[(x*3)+2] = image->format->palette->colors[pixels[x]].b;
  }
  *c=1; /* It's important that we know if dat is converted or not 
	   If it isn't it points to image->pixels and we cannot 
	   free that without SDL */
 } else
  dat = (unsigned char *)image->pixels;
 return dat;
}


/* 
 * Given a file name, return a SDL_Surface 
 */

SDL_Surface *file2surface( char *filename ) {
 SDL_Surface *image;
 image = IMG_Load(filename);
 if(!image) {
  e_push("IMG_Load");
  return (SDL_Surface *)0;
 }
 return image;

}

void vect_scale(vect_t *a, float *b) {
 *b = sqrt((a->x*a->x)+(a->y*a->y)+(a->z*a->z));
 return;
}

void norm_vect(vect_t *a) {
 float i,j; 
 vect_scale(a, &j);
 i = 1/j;
 a->x*=i;
 a->y*=i;
 a->z*=i;
 return;
}

void add_vect(vect_t *a, vect_t *b, vect_t *c) {
 c->x = a->x + b->x;
 c->y = a->y + b->y;
 c->z = a->z + b->z;
 return;
}

void sub_vect(vect_t *a, vect_t *b, vect_t *c) {
 c->x = a->x - b->x;
 c->y = a->y - b->y;
 c->z = a->z - b->z;
 return;
}

void mul_vect(vect_t *a, float *b, vect_t *c) {
 c->x = a->x * *b;
 c->y = a->y * *b;
 c->z = a->z * *b;
 return;
}

void div_vect(vect_t *a, float *b, vect_t *c) {
 c->x = a->x / *b;
 c->y = a->y / *b;
 c->z = a->z / *b;
 return;
}

void points2vect(vect_t *a, vect_t *b, vect_t *c) {
 c->x = a->x - b->x;
 c->y = a->y - b->y;
 c->z = a->z - b->z;
//  norm_vect(c);
 return;
}
