
 /*
		  ,----------------, 
	    ,-----------------------,
	  ,"                      ,"|
	 +-----------------------+  |
	 |  .-----------------.  |  |
	 |  |                 |  |  | 
	 |  | Hack the Gibson |  |  |
	 |- |      TERM       |  |  |
	 |- |                 |  |  | 
	 |- |                 |  |  | 
	 |- `-----------------'  |,"
	 +-----------------------+
	    /_)______________(_/
       _______________________________ 
      /  oooooooooooooooo  .o.  oooo /|
     / ==ooooooooooooooo==.o.  ooo= // 
    /_==__==========__==_ooo__ooo=_/'
    `-----------------------------'

   Hack the Gibson TERM is GNU software
    please read ../COPYING for details

 hgterm by Thea DeSilva
 */

 /*
  * vt100.c	ANSI/VT102 emulator code.
  *		This code was integrated to the Minicom communications
  *		package, but has been reworked to allow usage as a separate
  *		module.
  *
  *		It's not a "real" vt102 emulator - it's more than that:
  *		somewhere between vt220 and vt420 in commands.
  *
  *		This file is part of the minicom communications package,
  *		Copyright 1991-1995 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
  *		as published by the Free Software Foundation; either version
  *		2 of the License, or (at your option) any later version.
  *
  * // jl 04.09.97 character map conversions in and out
  *    jl 06.07.98 use conversion tables with the capture file
  *  Thea 06.22.04 mucked around a bit to merge into hgterm
  */
#include <sys/param.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <termcap.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
#include <SDL.h>
#include <SDL_image.h>

#include "include.h"

/* Begin local terminal commands */

 void _null(void) {
 }

 int wputc(WIN *win, char c, f_plane_t *plane) {
  #ifdef DEBUG
   printf("vt100->wputc( %c )\n", c);
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  f_char(c, plane);
 }

 int wputs( WIN *win, char *s, f_plane_t *plane) {
  #ifdef DEBUG
   printf("vt100->wputc( %s )\n", &s);
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  while(*s) wputc(win, *s++, plane);
 }

 /* Clear to end of line */
 int wclreol(WIN *win, f_plane_t *plane) {
  int i;
  #ifdef DEBUG
   printf("vt100->wclreol\n");
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  for(i = plane->window.curx; i < plane->x;i++)
   plane->data[ T2L(i, (plane->window.cury + plane->disp_start) % SCROLL_BUF, plane->x) ].c = 200;
 }

 /* Clear entire line */
 int wclrel(WIN *win, f_plane_t *plane) {
  int i;
  #ifdef DEBUG
   printf("vt100->wclrel\n");
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  for(i = 0; i<plane->x; i++)
   plane->data[ T2L( i, (plane->window.cury + plane->disp_start) % SCROLL_BUF, plane->x) ].c = 200;
 }

 /* Clear to beginning of screen */
 int wclrbos(WIN *win, f_plane_t *plane) {
  int x, y;
  #ifdef DEBUG
   printf("vt100->wclrbos\n");
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  x = plane->window.curx;
  for(y = plane->window.cury; y> 0; y--) {
   for(; x> 0; x--)
    plane->data[ T2L( x, (y + plane->disp_start) % SCROLL_BUF, plane->x) ].c = 200; 
   x = plane->x; 
  } 
 }

 /* Clear screen */
 int winclr(WIN *win, f_plane_t *plane) {
  int x, y;
  #ifdef DEBUG
   printf("vt100->winclr\n");
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  for(y = 0; y< plane->y; y++)
   for(x = 0;x< plane->x; x++) {
    plane->data[T2L( x, (y+plane->disp_start) % SCROLL_BUF, plane->x)].c = 200; 
   }
  
 }

 int wdelchar(WIN *w, f_plane_t *plane) {
  #ifdef DEBUG
   printf("vt100->wdelchar\n");
   printf("CURSOR(%d, %d)\n", w->curx, w->cury );
  #endif
  f_char(8, plane);
 }

 int winschar(WIN *win, f_plane_t *plane) {
  #ifdef DEBUG
   printf("vt100->winschar\n");
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  winschar2(win, ' ', 0, plane);
 }

 int wlocate(WIN *win, int x, int y, f_plane_t *plane) {
  #ifdef DEBUG
   printf("vt100->wlocate( %d, %d )\n", x, y);
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  if(x > 0 && x < plane->x) 
   win->curx = x;
  if(y > 0 && y < plane->y) 
   win->cury = y;
 }

 int wredraw() {
  #ifdef DEBUG
   printf("vt100->wredraw\n");
  #endif
 }

 int wcursor(WIN *w, int type) {
  #ifdef DEBUG
   printf("vt100->wcursor( %d )\n", type);
   printf("CURSOR(%d, %d)\n", w->curx, w->cury );
  #endif
  w->cursor = type;
 }


 int winschar2(WIN *w, int c, int move, f_plane_t *plane) {
  int x, y;
  int doit=1;
  f_char_t buf[560];
  f_char_t *e;
  int len, odir;
  int oldx;

  #ifdef DEBUG
   printf("vt100->winschar2( %d, %d )\n", c, move);
   printf("CURSOR(%d, %d)\n", w->curx, w->cury );
  #endif
  
  if(c == 0 || strchr("\r\n\t\b\007", c) || w->curx >= w->xs -1) {
   wputc( w, c, plane);
   return;
  }
  odir = w->direct;

  y = w->y1 + w->cury;
  x = w->x1 + w->curx;
  oldx = w->curx;
  len = w->xs - w->curx;
  memcpy(buf, &plane->data[ T2L(x, y, plane->x)], sizeof(f_char_t) * len);
  wputc(w, c, plane); 
  if(!move) w->curx = oldx;
  e = buf;
  for(++x; x<= w->x2; x++) {
   plane->data[ T2L( x, y, plane->x)].c == e->c;
   e++;
  }
  w->direct = odir;
  wlocate(w, w->curx, w->cury, plane);
 }

 int wscroll(WIN *win, int dir, f_plane_t *plane) {
  int x ,y; 
  void *a, *b, *c;
  unsigned char d;
 
  #ifdef DEBUG
   printf("vt100->wscroll( dir %d )\n");
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif

  if(dir == S_UP) {
   printf("s_up\n");
   for(y = plane->window.sy2; y>plane->window.sy1-1;y--)  {
    a = (f_char_t *)&plane->data[ plane->x * y     ];
    b = (f_char_t *)&plane->data[ plane->x * (y-1) ];
    if(y > plane->window.sy1) 
     memcpy(a, b, plane->x * sizeof(f_char_t) );
    else
     memset( a, 200, plane->x * sizeof(f_char_t) );
   }

  }
  
  if(dir == S_DOWN) {
   for(y = plane->window.sy1; y<plane->window.sy2+1;y++)  {
    a = (f_char_t *)&plane->data[ plane->x * y     ];
    b = (f_char_t *)&plane->data[ plane->x * (y+1) ];
    if(y < plane->window.sy2) 
     memcpy(a, b, plane->x * sizeof(f_char_t) );
    else
     memset( a, 200, plane->x * sizeof(f_char_t) );
   }
  }
   
 }

 /* Insert/delete */
 int winsline(WIN *w, f_plane_t *plane) {
  int osy1, osy2;

  #ifdef DEBUG
   printf("vt100->winsline\n");
   printf("CURSOR(%d, %d)\n", w->curx, w->cury );
  #endif
   
  osy1 = w->sy1;
  osy2 = w->sy2;

  w->sy1 = w->y1 + w->cury;
  w->sy2 = w->y2;
  if (w->sy1 < osy1) w->sy1 = osy1;
  if (w->sy2 > osy2) w->sy2 = osy2;
   wscroll(w, S_UP, plane);

  w->sy1 = osy1;
  w->sy2 = osy2;  
 }

 /* Delete line and scroll up?? */
 int wdelline(WIN *w, f_plane_t *plane) {
  int osy1, osy2;
  int ocury;
 
  #ifdef DEBUG
   printf("vt100->wdelline\n");
   printf("CURSOR(%d, %d)\n", w->curx, w->cury );
  #endif

  ocury = w->cury;
  osy1 = w->sy1;
  osy2 = w->sy2;

  w->sy1 = w->y1 + w->cury;
  w->sy2 = w->y2;
  if (w->sy1 < osy1) w->sy1 = osy1;
  if (w->sy2 > osy2) w->sy2 = osy2;
   wscroll(w, S_DOWN, plane);
  wlocate(w, 0, ocury, plane);

  w->sy1 = osy1;
  w->sy2 = osy2;
 }


 int wclrbol(WIN *win, f_plane_t *plane) {
  int x;

  #ifdef DEBUG 
   printf("vt100->wclrbol\n");
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  
  for(x = plane->window.curx; x> 0; x--) 
   plane->data[ T2L( x, (plane->window.cury+plane->disp_start) % SCROLL_BUF, plane->x) ].c = 200;
 }

 int wclreos(WIN * win, f_plane_t *plane) { 
  int x, y;

  #ifdef DEBUG
   printf("vt100->wclreos\n");
   printf("CURSOR(%d, %d)\n", win->curx, win->cury );
  #endif
  
  x = plane->window.curx;
  for(y = plane->window.cury; y< plane->y; y++) {
   for(; x< plane->x; x++) 
    plane->data[ T2L( x, (y+plane->disp_start) % SCROLL_BUF, plane->x) ].c = 200; 
   x = 0;
  }
 }


/* end local terminal commands */


#define OLD 0

 /*
  * The global variable esc_s holds the escape sequence status:
  * 0 - normal
  * 1 - ESC
  * 2 - ESC [
  * 3 - ESC [ ?
  * 4 - ESC (
  * 5 - ESC )
  * 6 - ESC #
  * 7 - ESC P
  */
 static int esc_s = 0;

#define ESC 27

 /* Structure to hold escape sequences. */
 struct escseq {
   int code;
   char *vt100_st;
   char *vt100_app;
   char *ansi;
 };

 /* Escape sequences for different terminal types. */
 static struct escseq vt_keys[] = {
   { K_F1,      "[17~", "[17~", "OP" },
   { K_F2,      "[18~", "[18~", "OQ" },
   { K_F3,      "[19~", "[19~", "OR" },
   { K_F4,      "[20~", "[20~", "OS" },
   { K_F5,      "[21~", "[21~", "OT" },
   { K_F6,      "[23~", "[23~", "OU" },
   { K_F7,      "[24~", "[24~", "OV" },
   { K_F8,      "[25~", "[25~", "OW" },
   { K_F9,      "[26~", "[26~", "OX" },
   { K_F10,     "[28~", "[28~", "OY" },
   { K_F11,     "[29~", "[29~", "OZ" },
   { K_F12,     "[31~", "[31~", "OA" },
   { K_HOME,    "OP",   "OP",   "[H" },
   { K_PGUP,    "OQ",   "OQ",   "[V" },
   { K_UP,      "[A",   "OA",   "[A" },
   { K_LT,      "[D",   "OD",   "[D" },
   { K_RT,      "[C",   "OC",   "[C" },
   { K_DN,      "[B",   "OB",   "[B" },
   { K_END,     "OR",   "OR",   "[Y" },
   { K_PGDN,    "OS",   "OS",   "[U" },
   { K_INS,     "[1~",  "[1~",  "[@"  },
   { K_DEL,     "[3~",  "[3~",  "\177" },
   { 0,         NULL,   NULL,   NULL }
 };

 /* Two tables for user-defined character map conversions.
  * defmap.h should contain all characters 0-255 in ascending order
  * to default to no conversion.    jl 04.09.1997
  */

 unsigned char vt_inmap[256] = {
#include "defmap.h"
 };
 unsigned char vt_outmap[256] = {
#include "defmap.h"
 };

 /*
  * Initialize the emulator once.
  */
 void vt_install(win, plane)
 WIN *win;
 f_plane_t *plane;
 {
   plane->vt100.vt_type = ANSI;
   plane->vt100.vt_wrap = 0;
   plane->vt100.vt_addlf = 0;
   plane->vt100.vt_asis = 0;
   plane->vt100.vt_bs = 8;
   plane->vt100.vt_insert = 0;
   plane->vt100.vt_crlf = 0;
   plane->vt100.ptr = 0;
   plane->vt100.newy1 = 0;
   plane->vt100.newy2 = 23;
   plane->vt100.savex = 0;
   plane->vt100.savey = 0;
   plane->vt100.saveattr = XA_NORMAL;
   plane->vt100.savecol = 112;
   plane->vt100.vt_win = NIL_WIN;
   plane->vt100.vt_keyb = &_null;
   plane->vt100.vt_win = win;
 }

 /* Partial init (after screen resize) */
 void vt_pinit(win, fg, bg, plane)
 WIN *win;
 int fg, bg;
 f_plane_t *plane;
 {
   plane->vt100.vt_win = win;
   plane->vt100.newy1 = 0;
   plane->vt100.newy2 = plane->vt100.vt_win->ys - 1;
   wresetregion(plane->vt100.vt_win);
   if (fg > 0) plane->vt100.vt_fg = fg;
   if (bg > 0) plane->vt100.vt_bg = bg;
   wsetfgcol(plane->vt100.vt_win, plane->vt100.vt_fg);
   wsetbgcol(plane->vt100.vt_win, plane->vt100.vt_bg);
 }

 /* Set characteristics of emulator. */
 void vt_init(type, fg, bg, wrap, add, plane)
 int type;
 int fg;
 int bg;
 int wrap;
 int add;
 f_plane_t *plane;
 {
   plane->vt100.vt_type = type;
   if (plane->vt100.vt_type == ANSI) {
	 plane->vt100.vt_fg = WHITE;
	 plane->vt100.vt_bg = BLACK;
   } else {
	 plane->vt100.vt_fg = fg;
	 plane->vt100.vt_bg = bg;
   }
   if (wrap >= 0) plane->vt100.vt_win->wrap = plane->vt100.vt_wrap = wrap;
   plane->vt100.vt_addlf = add;
   plane->vt100.vt_insert = 0;
   plane->vt100.vt_crlf = 0;
   plane->vt100.vt_om = 0;

   plane->vt100.newy1 = 0;
   plane->vt100.newy2 = plane->vt100.vt_win->ys - 1;
   wresetregion(plane->vt100.vt_win);
   plane->vt100.vt_keypad = NORMAL;
   plane->vt100.vt_cursor = NORMAL;
   plane->vt100.vt_tabs[0] = 0x01010100;
   plane->vt100.vt_tabs[1] =
   plane->vt100.vt_tabs[2] =
   plane->vt100.vt_tabs[3] =
   plane->vt100.vt_tabs[4] = 0x01010101;
   plane->vt100.ptr = 0;
   memset(plane->vt100.escparms, 0, sizeof(plane->vt100.escparms));
   esc_s = 0;

   if (plane->vt100.vt_keyb) (*plane->vt100.vt_keyb)(plane->vt100.vt_keypad, plane->vt100.vt_cursor);
   wsetfgcol(plane->vt100.vt_win, plane->vt100.vt_fg);
   wsetbgcol(plane->vt100.vt_win, plane->vt100.vt_bg);
 }

 /* Change some things on the fly. */
 void vt_set(addlf, wrap, capfp, docap, bscode, echo, cursor, asis, plane)
 int addlf;
 int wrap;
 FILE *capfp;
 int docap;
 int bscode;
 int echo;
 int cursor;
 int asis;
 f_plane_t *plane;
 {
   if (plane->vt100.vt_addlf >= 0) plane->vt100.vt_addlf = addlf;
   if (wrap >= 0)  plane->vt100.vt_win->wrap = plane->vt100.vt_wrap = wrap;
   if (capfp != (FILE *)0) plane->vt100.vt_capfp = capfp;
   if (docap >= 0) plane->vt100.vt_docap = docap;
   if (bscode >= 0) plane->vt100.vt_bs = bscode;
   if (cursor >= 0) plane->vt100.vt_cursor = cursor;
   if (asis >=0) plane->vt100.vt_asis = asis;
 }

 /* Output a string to the modem.  how do we take this? */
void v_termout(s, len, plane)
 char *s;
 int len;
 f_plane_t *plane;
 {
  int i;
  if(len > 0) {
   for(i=0;i<len;i++)
    write( qfd, &s[i], sizeof( unsigned char ) );
  } else {
   for(i = 0;i < strlen(s); i++)
    write( qfd, &s[i], sizeof( unsigned char) );
  }
 }

 /*
  * Escape code handling.
  */

 /*
  * ESC was seen the last time. Process the next character.
  */
 static void state1(c, plane)
 int c;
 f_plane_t *plane;
 {
   short x, y, f;

   switch(c) {
	 case '[': /* ESC [ */
		 esc_s = 2;
		 return;
	 case '(': /* ESC ( */
		 esc_s = 4;
		 return;
	 case ')': /* ESC ) */
		 esc_s = 5;
		 return;
	 case '#': /* ESC # */
		 esc_s = 6;
		 return;
	 case 'P': /* ESC P (DCS, Device Control String) */
		 esc_s = 7;
		 return;
	 case 'D': /* Cursor down */
	 case 'M': /* Cursor up */
		 x = plane->vt100.vt_win->curx;
		 if (c == 'D')  { /* Down. */
			 y = plane->vt100.vt_win->cury + 1;
			 if (y == plane->vt100.newy2 + 1)
				 wscroll(plane->vt100.vt_win, S_UP, plane);
			 else if (plane->vt100.vt_win->cury < plane->vt100.vt_win->ys)
				 wlocate(plane->vt100.vt_win, x, y, plane);
		 }		
		 if (c == 'M')  { /* Up. */
			 y = plane->vt100.vt_win->cury - 1;
			 if (y == plane->vt100.newy1 - 1)
				 wscroll(plane->vt100.vt_win, S_DOWN, plane);
			 else if (y >= 0)
				 wlocate(plane->vt100.vt_win, x, y, plane);
		 }
		 break;
	 case 'E': /* CR + NL */
		 wputs(plane->vt100.vt_win, "\r\n", plane);
		 break;
	 case '7': /* Save attributes and cursor position */
	 case 's':
		 plane->vt100.savex = plane->vt100.vt_win->curx;
		 plane->vt100.savey = plane->vt100.vt_win->cury;
		 plane->vt100.saveattr = plane->vt100.vt_win->attr;
		 plane->vt100.savecol = plane->vt100.vt_win->color;
		 break;
	 case '8': /* Restore them */
	 case 'u':	
		 plane->vt100.vt_win->color = plane->vt100.savecol; /* HACK should use wsetfgcol etc */
		 wsetattr(plane->vt100.vt_win, plane->vt100.saveattr);
		 wlocate(plane->vt100.vt_win, plane->vt100.savex, plane->vt100.savey, plane);
		 break;
	 case '=': /* Keypad into applications mode */
		 plane->vt100.vt_keypad = APPL;
		 if (plane->vt100.vt_keyb) (*plane->vt100.vt_keyb)(plane->vt100.vt_keypad, plane->vt100.vt_cursor);
		 break;
	 case '>': /* Keypad into numeric mode */
		 plane->vt100.vt_keypad = NORMAL;
		 if (plane->vt100.vt_keyb) (*plane->vt100.vt_keyb)(plane->vt100.vt_keypad, plane->vt100.vt_cursor);
		 break;
	 case 'Z': /* Report terminal type */
		 if (plane->vt100.vt_type == VT100)
			 v_termout("\033[?1;0c", 0);
		 else	
			 v_termout("\033[?c", 0);
		 break;	
	 case 'c': /* Reset to initial state */
		 f = XA_NORMAL;
		 wsetattr(plane->vt100.vt_win, f);
		 plane->vt100.vt_win->wrap = (plane->vt100.vt_type != VT100);
		 if (plane->vt100.vt_wrap != -1) plane->vt100.vt_win->wrap = plane->vt100.vt_wrap;
		 plane->vt100.vt_crlf = plane->vt100.vt_insert = 0;
		 vt_init(plane->vt100.vt_type, plane->vt100.vt_fg, plane->vt100.vt_bg, plane->vt100.vt_win->wrap, 0, plane);
		 wlocate(plane->vt100.vt_win, 0, 0, plane);
		 break;
	 case 'H': /* Set tab in current position */
		 x = plane->vt100.vt_win->curx;
		 if (x > 159) x = 159;
		 plane->vt100.vt_tabs[x / 32] |= 1 << (x % 32);
		 break;
	 case 'N': /* G2 character set for next character only*/
	 case 'O': /* G3 "				"    */
	 case '<': /* Exit vt52 mode */
	 default:
		 /* ALL IGNORED */
		 break;
   }
   esc_s = 0;
   return;
 }

 /* ESC [ ... [hl] seen. */
 static void ansi_mode(on_off, plane)
 int on_off;
 f_plane_t *plane;
 {
   int i;

   for(i = 0; i <= plane->vt100.ptr; i++) {
	 switch(plane->vt100.escparms[i]) {
		 case 4: /* Insert mode  */
			 plane->vt100.vt_insert = on_off;
			 break;
		 case 20: /* Return key mode */
			 plane->vt100.vt_crlf = on_off;
			 break;
	 }
   }
 }

 /*
  * ESC [ ... was seen the last time. Process next character.
  */
 static void state2(c, plane)
 int c;
 f_plane_t *plane;
 {
   short x, y, attr, f;
   char temp[32];

   /* See if a number follows */
   if (c >= '0' && c <= '9') {
	 plane->vt100.escparms[plane->vt100.ptr] = 10*(plane->vt100.escparms[plane->vt100.ptr]) + c - '0';
	 return;
   }
   /* Separation between numbers ? */
   if (c == ';') {
	 if (plane->vt100.ptr < 15) plane->vt100.ptr++;
	 return;
   }
   /* ESC [ ? sequence */
   if (plane->vt100.escparms[0] == 0 && plane->vt100.ptr == 0 && c == '?') {
	 esc_s = 3;
	 return;
   }

   /* Process functions with zero, one, two or more arguments */
   switch(c) {
	 case 'A':
	 case 'B':
	 case 'C':
	 case 'D': /* Cursor motion */
		 if ((f = plane->vt100.escparms[0]) == 0) f = 1;
		 x = plane->vt100.vt_win->curx;
		 y = plane->vt100.vt_win->cury;
		 x += f * ((c == 'C') - (c == 'D'));
		 if (x < 0) x = 0;
		 if (x >= plane->vt100.vt_win->xs) x = plane->vt100.vt_win->xs - 1;
		 if (c == 'B') { /* Down. */
			 y += f;
			 if (y >= plane->vt100.vt_win->ys) y = plane->vt100.vt_win->ys - 1;
			 if (y >= plane->vt100.newy2 + 1) y = plane->vt100.newy2;
		 }
		 if (c == 'A') { /* Up. */
			 y -= f;
			 if (y < 0) y = 0;
			 if (y <= plane->vt100.newy1 - 1) y = plane->vt100.newy1;
		 }	
		 wlocate(plane->vt100.vt_win, x, y, plane);
		 break;
	 case 'K': /* Line erasing */
		 switch(plane->vt100.escparms[0]) {
			 case 0:
				 wclreol(plane->vt100.vt_win,plane);
				 break;
			 case 1:
				 wclrbol(plane->vt100.vt_win,plane);
				 break;
			 case 2:
				 wclrel(plane->vt100.vt_win,plane);
				 break;
		 }
		 break;
	 case 'J': /* Screen erasing */
		 x = plane->vt100.vt_win->color;
		 y = plane->vt100.vt_win->attr;
		 if (plane->vt100.vt_type == ANSI) {
			 wsetattr(plane->vt100.vt_win, XA_NORMAL);
			 wsetfgcol(plane->vt100.vt_win, WHITE);
			 wsetbgcol(plane->vt100.vt_win, BLACK);
		 }
		 switch(plane->vt100.escparms[0]) {
			 case 0:
				 wclreos(plane->vt100.vt_win, plane);
				 break;
			 case 1:
				 wclrbos(plane->vt100.vt_win,plane);
				 break;
			 case 2:
				 winclr(plane->vt100.vt_win,plane);
				 break;
		 }
		 if (plane->vt100.vt_type == ANSI) {
			 plane->vt100.vt_win->color = x;
			 plane->vt100.vt_win->attr = y;
		 }
		 break;
	 case 'n': /* Requests / Reports */
		 switch(plane->vt100.escparms[0]) {
			 case 5: /* Status */
				 v_termout("\033[0n", 0);
				 break;
			 case 6:	/* Cursor Position */
				 sprintf(temp, "\033[%d;%dR", 
					 plane->vt100.vt_win->cury + 1, plane->vt100.vt_win->curx + 1);
				 v_termout(temp, 0);
				 break;
		 }
		 break;
	 case 'c': /* Identify Terminal Type */
		 if (plane->vt100.vt_type == VT100) {
			 v_termout("\033[?1;2c", 0);
			 break;
		 }
		 v_termout("\033[?c", 0);
		 break;
	 case 'x': /* Request terminal parameters. */
		   /* Always answers 19200-8N1 no options. */
		 sprintf(temp, "\033[%c;1;1;120;120;1;0x",
			 plane->vt100.escparms[0] == 1 ? '3' : '2');
		 v_termout(temp, 0);
		 break;
	 case 's': /* Save attributes and cursor position */
		 plane->vt100.savex = plane->vt100.vt_win->curx;
		 plane->vt100.savey = plane->vt100.vt_win->cury;
		 plane->vt100.saveattr = plane->vt100.vt_win->attr;
		 plane->vt100.savecol = plane->vt100.vt_win->color;
		 break;
	 case 'u': /* Restore them */
		 break;
		 plane->vt100.vt_win->color = plane->vt100.savecol; /* HACK should use wsetfgcol etc */
		 wsetattr(plane->vt100.vt_win, plane->vt100.saveattr);
		 wlocate(plane->vt100.vt_win, plane->vt100.savex, plane->vt100.savey, plane);
		 break;
	 case 'h':
		 ansi_mode(1);
		 break;
	 case 'l':
		 ansi_mode(0);
		 break;
	 case 'H':
	 case 'f': /* Set cursor position */
		 if ((y = plane->vt100.escparms[0]) == 0) y = 1;
		 if ((x = plane->vt100.escparms[1]) == 0) x = 1;
		 if( x == 1 && y == 1) {
		  plane->window.curx = 0;
		  plane->window.cury = 0;
		  break;
		 }
		 if (plane->vt100.vt_om) y += plane->vt100.newy1;
		 wlocate(plane->vt100.vt_win, x - 1, y - 1, plane);
		 break;
	 case 'g': /* Clear tab stop(s) */
		 if (plane->vt100.escparms[0] == 0) {
			 x = plane->vt100.vt_win->curx;
			 if (x > 159) x = 159;
			 plane->vt100.vt_tabs[x / 32] &= ~ (1 << x % 32);
		 }
		 if (plane->vt100.escparms[0] == 3)
			 for(x = 0; x < 5; x++) plane->vt100.vt_tabs[x] = 0;
		 break;
	 case 'm': /* Set attributes */
		 attr = wgetattr((plane->vt100.vt_win));
		 for (f = 0; f <= plane->vt100.ptr; f++) {
		     if (plane->vt100.escparms[f] >= 30 && plane->vt100.escparms[f] <= 37)
			 wsetfgcol(plane->vt100.vt_win, plane->vt100.escparms[f] - 30);
		     if (plane->vt100.escparms[f] >= 40 && plane->vt100.escparms[f] <= 47)
			 wsetbgcol(plane->vt100.vt_win, plane->vt100.escparms[f] - 40);
		     switch(plane->vt100.escparms[f]) {
			 case 0:
				 attr = XA_NORMAL;
				 wsetfgcol(plane->vt100.vt_win, plane->vt100.vt_fg);
				 wsetbgcol(plane->vt100.vt_win, plane->vt100.vt_bg);
				 break;
			 case 4:
				 attr |= XA_UNDERLINE;
				 break;
			 case 7:
				 attr |= XA_REVERSE;
				 break;
			 case 1:
				 attr |= XA_BOLD;
				 break;
			 case 5:
				 attr |= XA_BLINK;
				 break;
			 case 22: /* Bold off */
				 attr &= ~XA_BOLD;
				 break;
			 case 24: /* Not underlined */
				 attr &=~XA_UNDERLINE;
				 break;
			 case 25: /* Not blinking */
				 attr &= ~XA_BLINK;
				 break;
			 case 27: /* Not reverse */
				 attr &= ~XA_REVERSE;
				 break;
			 case 39: /* Default fg color */
				 wsetfgcol(plane->vt100.vt_win, plane->vt100.vt_fg);
				 break;
			 case 49: /* Default bg color */
				 wsetbgcol(plane->vt100.vt_win, plane->vt100.vt_bg);
				 break;
				 
		     }
		 }
		 wsetattr(plane->vt100.vt_win, attr);
		 break;
	 case 'L': /* Insert lines */
		 if ((x = plane->vt100.escparms[0]) == 0) x = 1;
		 for(f = 0; f < x; f++)
			 winsline(plane->vt100.vt_win,plane);
		 break;	
	 case 'M': /* Delete lines */
		 if ((x = plane->vt100.escparms[0]) == 0) x = 1;
		 for(f = 0; f < x; f++)
			 wdelline(plane->vt100.vt_win,plane);
		 break;	
	 case 'P': /* Delete Characters */
		 if ((x = plane->vt100.escparms[0]) == 0) x = 1;
		 for(f = 0; f < x; f++)
			 wdelchar(plane->vt100.vt_win,plane);
		 break;	
	 case '@': /* Insert Characters */		
		 if ((x = plane->vt100.escparms[0]) == 0) x = 1;
		 for(f = 0; f < x; f++)
			 winschar(plane->vt100.vt_win, plane);
		 break;	
	 case 'r': /* Set scroll region */
		 if ((plane->vt100.newy1 = plane->vt100.escparms[0]) == 0) plane->vt100.newy1 = 1;
		 if ((plane->vt100.newy2 = plane->vt100.escparms[1]) == 0) plane->vt100.newy2 = plane->vt100.vt_win->ys;
		 plane->vt100.newy1-- ; plane->vt100.newy2--;
		 if (plane->vt100.newy1 < 0) plane->vt100.newy1 = 0;
		 if (plane->vt100.newy2 < 0) plane->vt100.newy2 = 0;
		 if (plane->vt100.newy1 >= plane->vt100.vt_win->ys) plane->vt100.newy1 = plane->vt100.vt_win->ys - 1;
		 if (plane->vt100.newy2 >= plane->vt100.vt_win->ys) plane->vt100.newy2 = plane->vt100.vt_win->ys - 1;
		 if (plane->vt100.newy1 >= plane->vt100.newy2) {
			 plane->vt100.newy1 = 0;
			 plane->vt100.newy2 = plane->vt100.vt_win->ys - 1;
		 }
		 wsetregion(plane->vt100.vt_win, plane->vt100.newy1, plane->vt100.newy2);
 // 		printf("region %d %d; %d %d\n", plane->window.sy1, plane->window.sy2, plane->vt100.newy1, plane->vt100.newy2);
		 wlocate(plane->vt100.vt_win, 0, plane->vt100.newy1, plane);
		 break;
	 case 'i': /* Printing */
	 case 'y': /* Self test modes */
	 default:
		 /* IGNORED */
		 break;
   }
   /* Ok, our escape sequence is all done */
   esc_s = 0;
   plane->vt100.ptr = 0;
   memset(plane->vt100.escparms, 0, sizeof(plane->vt100.escparms));
   return;		
 }
   
 /* ESC [? ... [hl] seen. */
 static void dec_mode(on_off, plane)
 int on_off;
 f_plane_t *plane;
 {
   int i;

   for(i = 0; i <= plane->vt100.ptr; i++) {
	 switch(plane->vt100.escparms[i]) {
		 case 1: /* Cursor keys in cursor/appl mode */
			 plane->vt100.vt_cursor = on_off ? APPL : NORMAL;
			 if (plane->vt100.vt_keyb) (*plane->vt100.vt_keyb)(plane->vt100.vt_keypad, plane->vt100.vt_cursor);
			 break;
		 case 6: /* Origin mode. */
			 plane->vt100.vt_om = on_off;
			 wlocate(plane->vt100.vt_win, 0, plane->vt100.newy1, plane);
			 break;
		 case 7: /* Auto wrap */
			 plane->vt100.vt_win->wrap = on_off;
			 break;
		 case 25: /* Cursor on/off */
			 wcursor(plane->vt100.vt_win, on_off ? CNORMAL : CNONE);
			 break;
		 case 67: /* Backspace key sends. (FIXME: vt420) */
			 /* setbackspace(on_off ? 8 : 127); */
			 break;
		 default: /* Mostly set up functions */
			 /* IGNORED */
			 break;
	 }
   }
 }

 /*
  * ESC [ ? ... seen.
  */
 static void state3(c, plane)
 int c;
 f_plane_t *plane;
 {
   /* See if a number follows */
   if (c >= '0' && c <= '9') {
	 plane->vt100.escparms[plane->vt100.ptr] = 10*(plane->vt100.escparms[plane->vt100.ptr]) + c - '0';
	 return;
   }
   switch(c) {
    /* It's interesting that somethine would crash here */
	 case 'h':
		 dec_mode(1);
		 break;
	 case 'l':
		 dec_mode(0);
		 break;
	 case 'i': /* Printing */
	 case 'n': /* Request printer status */
	 default:
		 /* IGNORED */
		 break;
   }
   esc_s = 0;
   plane->vt100.ptr = 0;
   memset(plane->vt100.escparms, 0, sizeof(plane->vt100.escparms));
   return;
 }

 /*
  * ESC ( Seen.
  */
 /*ARGSUSED*/
 static void state4(c, plane)
 int c;
 f_plane_t *plane;
 {
   /* Switch Character Sets. */
   (void)c;
   esc_s = 0;
 }

 /*
  * ESC ) Seen.
  */
 /*ARGSUSED*/
 static void state5(c, plane)
 int c;
 f_plane_t *plane;
 {
   /* Switch Character Sets. */
   (void)c;
   esc_s = 0;
 }

 /*
  * ESC # Seen.
  */
 /*ARGSUSED*/
 static void state6(c, plane)
 int c;
 f_plane_t *plane;
 {
   int x, y;

   /* Double height, double width and selftests. */
   switch(c) {
	 case '8':
		 /* Selftest: fill screen with E's */
		 plane->vt100.vt_win->doscroll = 0;
		 plane->vt100.vt_win->direct = 0;
		 wlocate(plane->vt100.vt_win, 0, 0, plane);
		 for(y = 0; y < plane->vt100.vt_win->ys; y++) {
			 wlocate(plane->vt100.vt_win, 0, y, plane);
			 for(x = 0; x < plane->vt100.vt_win->xs; x++)
				 wputc(plane->vt100.vt_win, 'E', plane);
		 }
		 wlocate(plane->vt100.vt_win, 0, 0, plane);
		 plane->vt100.vt_win->doscroll = 1;
		 wredraw(plane->vt100.vt_win, 1);
		 break;
	 default:
		 /* IGNORED */
		 break;
   }
   esc_s = 0;
 }

 /*
  * ESC P Seen.
  */
 static void state7(c, plane)
 int c;
 f_plane_t *plane;
 {
   /*
    * Device dependant control strings. The Minix virtual console package
    * uses these sequences. We can only turn cursor on or off, because
    * that's the only one supported in termcap. The rest is ignored.
    */
   static char buf[17];
   static int pos = 0;
   static int state = 0;

   if (c == ESC) {
	 state = 1;
	 return;
   }
   if (state == 1) {
	 buf[pos] = 0;
	 pos = 0;
	 state = 0;
	 esc_s = 0;
	 if (c != '\\') return;
	 /* Process string here! */
	 if (!strcmp(buf, "cursor.on")) wcursor(plane->vt100.vt_win, CNORMAL);
	 if (!strcmp(buf, "cursor.off")) wcursor(plane->vt100.vt_win, CNONE);
	 if (!strcmp(buf, "linewrap.on")) {
		 plane->vt100.vt_wrap = -1;
		 plane->vt100.vt_win->wrap = 1;
	 }
	 if (!strcmp(buf, "linewrap.off")) {
		 plane->vt100.vt_wrap = -1;
		 plane->vt100.vt_win->wrap = 0;
	 }
	 return;
   }
   if (pos > 15) return;
   buf[pos++] = c;
 }

 void vt_out(ch, plane)
 int ch;
 f_plane_t *plane;
 {
   m_ent_t *obj[2];
   int f;
   unsigned char c;
   int go_on = 0;

   
   if (!ch) return;

   c = (unsigned char)ch;
 
   
   
   if (plane->vt100.vt_docap == 2) /* Literal. */
	 fputc(c, plane->vt100.vt_capfp);

   /* Process <31 chars first, even in an escape sequence. */
   switch(c) {
	 case 5: /* AnswerBack for vt100's */
		 if (plane->vt100.vt_type != VT100) {
			 go_on = 1;
			 break;
		 }
		 v_termout("@(#)hgterm", 0);
		 break;
	 case '\r': /* Carriage return */
		 wputc(plane->vt100.vt_win, c, plane);
		 if (plane->vt100.vt_addlf) {
			 wputc(plane->vt100.vt_win, '\n', plane);
			 if (plane->vt100.vt_docap == 1)
				 fputc('\n', plane->vt100.vt_capfp);
		 }
		 break;
	  case '\t': /* Non - destructive TAB */
		 /* Find next tab stop. */
		 for(f = plane->vt100.vt_win->curx + 1; f < 160; f++)
		      if (plane->vt100.vt_tabs[f / 32] & (1 << f % 32)) break;
		 if (f >= plane->vt100.vt_win->xs) f = plane->vt100.vt_win->xs - 1;
			 wlocate(plane->vt100.vt_win, f, plane->vt100.vt_win->cury, plane);
		 if (plane->vt100.vt_docap == 1) fputc(c, plane->vt100.vt_capfp);
		 break;
	 case 013: /* Old Minix: CTRL-K = up */
		 wlocate(plane->vt100.vt_win, plane->vt100.vt_win->curx, plane->vt100.vt_win->cury - 1, plane);
		 break;
	 case '\f': /* Form feed: clear screen. */
		 winclr(plane->vt100.vt_win,plane);
		 wlocate(plane->vt100.vt_win, 0, 0, plane);
		 break;
	 case 14:
	 case 15:  /* Change character set. Not supported. */
		 break;
	 case 24:
	 case 26:  /* Cancel escape sequence. */
		 esc_s = 0;
		 break;
	 case ESC: /* Begin escape sequence */
		 esc_s = 1;
		 break;
	 case 128+ESC: /* Begin ESC [ sequence. */
		 esc_s = 2;
		 break;
	 case '\n':
	 case '\b':
		 f_char(c, plane);
		 break;
	 case 7: /* Bell */
		 wputc(plane->vt100.vt_win, c, plane);
		 if (plane->vt100.vt_docap == 1)
			 fputc(c, plane->vt100.vt_capfp);
		 break;
	 default:
		 go_on = 1;
		 break;
   }
   if (!go_on) return;

   /* Now see which state we are in. */
   switch(esc_s) {
	 case 0: /* Normal character */
		 c = vt_inmap[c];    /* conversion 04.09.97 / jl */
		 if (plane->vt100.vt_insert)
			 winschar2(plane->vt100.vt_win, c, 1, plane);
		 else
			 wputc(plane->vt100.vt_win, c, plane);
		 break;
	 case 1: /* ESC seen */
		 state1(c);
		 break;
	 case 2: /* ESC [ ... seen */
		 state2(c);
		 break;
	 case 3:
		 state3(c);
		 break;
	 case 4:
		 state4(c);
		 break;
	 case 5:
		 state5(c);
		 break;
	 case 6:
		 state6(c);
		 break;
	 case 7:
		 state7(c);
		 break;	
   }
 }

 /* Translate keycode to escape sequence. */
 void vt_send(c, plane)
 int c;
 f_plane_t *plane;
 {
   char s[3];
   int f;
   int len = 1;

   /* Special key? */
   if (c < 256) {
	 /* Translate backspace key? */
	 if (c == K_ERA) c = plane->vt100.vt_bs;
	 s[0] = vt_outmap[(unsigned char) c];  /* conversion 04.09.97 / jl */
	s[1] = 0;
	/* CR/LF mode? */
	if (c == '\r' && plane->vt100.vt_crlf) {
		s[1] = '\n';
		s[2] = 0;
		len = 2;
	}
	v_termout(s, len);
	if (plane->vt100.vt_nl_delay > 0 && c == '\r')
		usleep(1000 * plane->vt100.vt_nl_delay);
	printf( "alway, girl. Always\n");
	return;
  }

  /* Look up code in translation table. */
  for(f = 0; vt_keys[f].code; f++)
	if (vt_keys[f].code == c) break;
  if (vt_keys[f].code == 0) return;

  /* Now send appropriate escape code. */
  v_termout("\033", 0);
  if (plane->vt100.vt_type == VT100) {
	if (plane->vt100.vt_cursor == NORMAL)
		v_termout(vt_keys[f].vt100_st, 0);
	else
		v_termout(vt_keys[f].vt100_app, 0);
  } else
	v_termout(vt_keys[f].ansi, 0);
}
