/*
                 ,----------------, 
           ,-----------------------,
         ,"                      ,"|
        +-----------------------+  |
        |  .-----------------.  |  |
        |  |                 |  |  | 
        |  | Hack the Gibson |  |  |
        |- |      TERM       |  |  |
        |- |                 |  |  | 
        |- |                 |  |  | 
        |- `-----------------'  |,"
        +-----------------------+
           /_)______________(_/
      _______________________________ 
     /  oooooooooooooooo  .o.  oooo /|
    / ==ooooooooooooooo==.o.  ooo= // 
   /_==__==========__==_ooo__ooo=_/'
   `-----------------------------'

  Hack the Gibson TERM is GNU software
   please read ../COPYING for details

hgterm by Thea DeSilva
*/

/*
 * vt100.h	Header file for the vt100 emulator.
 *
 *		This file is part of the minicom communications package,
 *		Copyright 1991-1995 Miquel van Smoorenburg.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

typedef struct _elm {
  char value;
  char attr;
  char color;
} ELM;

/*
 * Control struct of a window
 */
typedef struct _win {
  int x1, y1, x2, y2;	/* Coordinates */
  int sy1, sy2;		/* Scrolling region */
  int xs, ys;		/* x and y size */
  char border;		/* type of border */
  char cursor;		/* Does it have a cursor */
  char attr;		/* Current attribute of window */
  char color;		/* Current color of window */
  char autocr;		/* With '\n', do an automatic '\r' */
  char doscroll;	/* Automatically scroll up */
  char wrap;		/* Wrap around edge */
  char direct;		/* Direct write to screen ? */
  short curx, cury;	/* current x and y */
  char o_cursor;
  short o_curx;
  short o_cury;
  char o_attr;
  char o_color;		/* Position & attributes before window was opened */
  ELM *map;		/* Map of contents */
  ELM *histbuf;		/* History buffer. */
  int histlines;	/* How many lines we keep in the history buffer */
  int histline;		/* Current line in the history buffer. */
} WIN;

typedef struct {
 int vt_nl_delay;                /* Delay after CR key */
 int vt_type;             	 /* Terminal type. */
 int vt_wrap;        	 	 /* Line wrap on/off */
 int vt_addlf;       	 	 /* Add linefeed on/off */
 int vt_fg;               	 /* Standard foreground color. */
 int vt_bg;               	 /* Standard background color. */
 int vt_keypad;           	 /* Keypad mode. */
 int vt_cursor;           	 /* cursor key mode. */
 int vt_asis;         	 	 /* 8bit clean mode. */
 int vt_bs;           	 	 /* Code that backspace key sends. */
 int vt_insert;       	 	 /* Insert mode */
 int vt_crlf;         	 	 /* Return sends CR/LF */
 int vt_om;               	 /* Origin mode. */
 WIN *vt_win; 			 /* Output window. */
 int vt_docap;            	 /* Capture on/off. */
 FILE *vt_capfp;          	 /* Capture file. */
 void (*vt_keyb)();       	 /* Gets called for NORMAL/APPL switch. */
 
 int escparms[8];         	 /* Cumulated escape sequence. */
 int ptr;             	 	 /* Index into escparms array. */
 long vt_tabs[5];         	 /* Tab stops for max. 32*5 = 160 columns. */
 
 short newy1;                	 /* Current size of scrolling region. */
 short newy2;
 
 /* Saved color and posistions */
 short savex, savey, saveattr, savecol;

} vt100_t; 

#define XA_NORMAL	 0 
#define XA_BLINK		 1
#define XA_BOLD		 2
#define XA_REVERSE	 4
#define XA_STANDOUT	 8
#define XA_UNDERLINE	16
#define XA_ALTCHARSET	32
#define XA_BLANK		64

/*
 * Possible colors
 */
#define BLACK		0
#define RED		1
#define GREEN		2
#define YELLOW		3
#define BLUE		4
#define MAGENTA		5
#define CYAN		6
#define WHITE		7

#define COLATTR(fg, bg) (((fg) << 4) + (bg))
#define COLFG(ca)	((ca) >> 4)
#define COLBG(ca)	((ca) % 16)

/*
 * Possible borders.
 */
#define BNONE	0
#define BSINGLE 1
#define BDOUBLE 2

/*
 * Scrolling directions.
 */
#define S_UP	1
#define S_DOWN	2

/*
 * Cursor types.
 */
#define CNONE	0
#define CNORMAL	1

/*
 * Title Positions
 */
#define TLEFT	0
#define TMID	1
#define TRIGHT	2

/*
 * For wselect.
 */
#define NIL_FUNLIST	((void (**)()) 0)
#define NIL_FUN		((void (*)()) 0)
#define MENU_END	((char *) 0)

#define CNULL		((char *)0)
#define NIL_WIN         (WIN *) 0

/*
 * Some macro's that can be used as functions.
 */
#define wsetregion(w, z1, z2) (((w)->sy1=(w)->y1+(z1)),((w)->sy2=(w)->y1+(z2)))
#define wresetregion(w) ( (w)->sy1 = (w)->y1, (w)->sy2 = (w)->y2 )
#define wgetattr(w) ( (w)->attr )
#define wsetattr(w, a) ( (w)->attr = (a) )
#define wsetfgcol(w, fg) ( (w)->color = ((w)->color & 15) + ((fg) << 4))
#define wsetbgcol(w, bg) ( (w)->color = ((w)->color & 240) + (bg) )
#define wsetam(w) ( (w)->wrap = 1 )
#define wresetam(w) ( (w)->wrap = 0 )

/*
 * Allright, now the macro's for our keyboard routines.
 */

#define K_BS		8
#define K_ESC		27
#define K_STOP		256
#define K_F1		257
#define K_F2		258
#define K_F3		259
#define K_F4		260
#define K_F5		261
#define K_F6		262
#define K_F7		263
#define K_F8		264
#define K_F9		265
#define K_F10		266
#define K_F11		277
#define K_F12		278
#define K_HOME		267
#define K_PGUP		268
#define K_UP		269
#define K_LT		270
#define K_RT		271
#define K_DN		272
#define K_END		273
#define K_PGDN		274
#define K_INS		275
#define K_DEL		276

#define NUM_KEYS	23
#define KEY_OFFS	256

/* Here's where the meta keycode start. (512 + keycode). */
#define K_META		512

#define K_ERA		'\b'
#define K_KILL		((int) -2)

/* Internal structure. */
struct key {
	char *cap;
	char len;
};


#define NORMAL	1
#define APPL	2

/* Don't change - hardcoded in minicom's dial.c */
#define VT100	1
#define ANSI	3


/* Prototypes from vt100.c */
/*
void vt_install(WIN *, f_plane_t *);
void vt_init(int, int, int, int, int, f_plane_t *);
void vt_pinit(WIN *, int, int, f_plane_t *);
void vt_set(int, int, FILE *, int, int, int, int, int, f_plane_t *);
void vt_out(int, f_plane_t *);
void vt_send(int, f_plane_t *);
*/
